##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: interfaces.py 82223 2007-12-10 12:59:59Z schwendinger $
"""
__docformat__ = 'restructuredtext'

from zope import interface
from zope import schema


class IMemCachedSessionDataContainer(interface.Interface):
    """A session container using memcache"""

    cacheName = schema.TextLine(title=u'Cachename',
                                required=False,
                                default=u'')




##############################################################################
#
# Copyright (c) 2008 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup for localy.session package

$Id: setup.py 89087 2008-07-31 14:13:31Z srichter $
"""
import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(name='lovely.session',
    version = '0.1.3',
    author = "Lovely Systems GmbH",
    author_email = "office@lovelysystems.com",
    description = "memcache-based session storage",
    long_description=(
        read('src', 'lovely', 'session', 'README.txt')
        + '\n\n' +
        read('CHANGES.txt')
        + '\n\n' +
        '========\n'
        'Download\n'
        '========\n'
        ),
    license='ZPL 2.1',
    keywords = "zope3 session memcache",
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Framework :: Zope3'],
    url='http://cheeseshop.python.org/pypi/lovely.session',
    packages=find_packages('src'),
    package_dir = {'': 'src'},
    namespace_packages=['lovely',],
    extras_require = dict(
        test = ['zope.app.testing',
                'zope.interface',
                'zope.security',
                'zope.testing',]),
    install_requires = [
        'setuptools',
        'lovely.memcached',
        'zope.app.container',
        'zope.session',
        'zope.schema',
        ],
    extras_require = dict(
        test=['zope.app.testing',
              'zope.app.zptpage',
              'zope.app.securitypolicy',
              'zope.app.zcmlfiles']),
    include_package_data = True,
    zip_safe = False,
    )



##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: tests.py 82223 2007-12-10 12:59:59Z schwendinger $
"""
__docformat__ = 'restructuredtext'

import unittest
from zope import component
import zope.interface
import zope.security
from zope.testing import doctest
from zope.testing.doctestunit import DocTestSuite, DocFileSuite
from zope.app.testing import setup


def test_suite():
    level2Suites = (
        DocFileSuite('README.txt',
             optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
             ),
            )
    for suite in level2Suites:
        suite.level = 2
    return unittest.TestSuite(level2Suites)

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2007 Lovely Systems and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: memcached.py 89069 2008-07-30 22:22:56Z srichter $
"""
__docformat__ = 'restructuredtext'
import md5
import cPickle
import memcache
import time
from threading import local

import persistent
import transaction
from transaction.interfaces import IDataManager

from zope import interface
from zope import component

from zope.schema.fieldproperty import FieldProperty

from zope.app.container.contained import Contained
from zope.session.interfaces import (
    ISessionDataContainer,
    ISessionData,
    ISessionPkgData,
    )
from zope.session.session import PersistentSessionDataContainer

from lovely.memcached.interfaces import IMemcachedClient
from lovely.session.interfaces import IMemCachedSessionDataContainer


class MemCachedSessionDataContainer(persistent.Persistent, Contained):
    interface.implements(ISessionDataContainer, IMemCachedSessionDataContainer)

    cacheName = FieldProperty(IMemCachedSessionDataContainer['cacheName'])

    def __init__(self):
        self.resolution = 5*60
        self.timeout = 1 * 60 * 60

    def get(self, key, default=None):
        return self[key]

    def __contains__(self, key):
        return key in self.dataManagers

    def __getitem__(self, key):
        # getitem never fails to make sure the session uses our own PkgData
        # implementation.
        if key in self.dataManagers:
            dm = self.dataManagers[key]
        else:
            dm = DataManager(self, key)
            txn = transaction.manager.get()
            txn.join(dm)
            self.dataManagers[key] = dm
            item = self.client.query(self._buildKey(key))
            if item is None:
                item = MemCacheSessionData()
            dm.data = item
        return dm.data

    def __setitem__(self, key, value):
        pass

    def items(self):
        return self.dataManagers.items()

    def _commitManager(self, key):
        if key in self.dataManagers:
            dm = self.dataManagers[key]
            if not dm.changed:
                del self.dataManagers[key]
                return
            if self.client.set(dm.data,
                               self._buildKey(key),
                               self.timeout):
                del self.dataManagers[key]

    def _abortManager(self, key):
        #TODO: ???
        if key in self.dataManagers:
            del self.dataManagers[key]

    def _buildKey(self, key):
        m = md5.new(key)
        return 'lovely.session.memcached:%s/%s'%(
                    self.__name__.encode('utf-8'),
                    m.hexdigest())

    @property
    def dataManagers(self):
        res = getattr(self.storage, 'dataManagers', None)
        if res is None:
            res = {}
            self.storage.dataManagers = res
        return res

    @property
    def client(self):
        return component.getUtility(IMemcachedClient, self.cacheName)

    @property
    def storage(self):
        # we use a thread local storage to have a memcache client for every
        # thread.
        if not hasattr(self, '_v_storage'):
            self._v_storage = local()
        return self._v_storage


class MemCacheData(dict):
    """See zope.app.session.interfaces.ISessionData
    """
    interface.implements(ISessionPkgData)


class MemCacheSessionData(dict):
    """See zope.app.session.interfaces.ISessionData"""
    interface.implements(ISessionData)
    lastAccessTime = 0

    def __init__(self):
        self.lastAccessTime = int(time.time())

    def __getitem__(self, key):
        try:
            return super(MemCacheSessionData, self).__getitem__(key)
        except KeyError:
            data = MemCachePkgData()
            self[key] = data
            return data


class MemCachePkgData(dict):
    """"""
    interface.implements(ISessionPkgData)


class DataManager(object):
    """A data manager for the transaction management of the session data"""
    interface.implements(IDataManager)

    def __init__(self, sessionData, key):
        self.sessionData = sessionData
        self._data = None
        self.key = key

    @apply
    def data():
        def get(self):
            return self._data
        def set(self, value):
            self._data = value
            self._oldData = cPickle.dumps(value)
        return property(get, set)

    @property
    def changed(self):
        return cPickle.dumps(self.data)!=self._oldData

    def abort(self, trans):
        self.sessionData._abortManager(self.key)

    def tpc_begin(self, trans):
        pass

    def commit(self, trans):
        pass

    def tpc_vote(self, trans):
        pass

    def tpc_finish(self, trans):
        self.sessionData._commitManager(self.key)

    def tpc_abort(self, trans):
        self.sessionData._abortManager(self.key)

    def sortKey(self):
        return str(id(self))

